/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: compile.c,v 1.70 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dlfcn.h>
#include "lang.h"
#include "misc.h"
#include "snd.h"

	/* the maximum number of local temporaries to allocate */
#define MAX_LOCALS 64
#define NUM_COMPILE_ARGS 8		/* must match 'value_t v1,...' line,
					 * and in lang.h, lang.c, and eval.c */


static char *src_fn = NULL;	/* filled in with full path of the .c file */
static FILE *curr_out = NULL;



static char *run_compiler(void);
static struct compiled_rec *load_dll(struct tree_rec *tr, char *fn, char *hash);
static void compile_prol(struct tree_rec *r, int *num_temps, int *num_itemps);
static void compile_epil(char *done, char *hash, int num_temps, int num_itemps);
static void compile_tree(struct tree_rec *tr, int local, int *num_temps, int *num_itemps, char **done, int depth);
static void compile_top_tree(struct tree_rec *tr, int *num_temps, int *num_itemps, char **done);
static void line_push(char *s, int depth);
static void line_append(char *s);
static void line_pop(void);
static char *cache_hash(struct tree_rec *tr);
static struct compiled_rec *cache_search(char *hash, struct tree_rec *tr);
static void cache_save(char *fn, char *hash);
static char *value_out(value_t val);

static char *done0 = "0";		/* used as argument to done_*() */
static char *done1 = "1";		/* used as argument to done_*() */
static char *done_and(char *d,char *e);
static char *done_or(char *d,char *e);

static char *
tmp_path(void)
{
	char *e;
	if (cache_dir) {
		return cache_dir;
	}
	e = getenv("TMPDIR");
	if (e) {
		return e;
	}
	return "/tmp";
}


struct compiled_rec *
compile_func(struct tree_rec *tr)
{
	int num_temps = 0;
	int num_itemps = 0;
	int save_num_errors = num_errors;
	struct compiled_rec *ret;
	char *done;
	char *dll_fn;
	char *hash;
	struct tree_rec *val;

	/* the optimizer, among other things, puts a flatted out tree in
	 * tr->u.fn_def.opt_expr. */
	opt(tr);

	hash = cache_hash(tr->u.fn_def.opt_expr);

	if ((ret = cache_search(hash, tr))) {
		return ret;
	}

	num_errors = 0;

	if (!src_fn) {
		char *path = tmp_path();
		src_fn = safe_malloc(strlen(path)+40);
		sprintf(src_fn,"%s/ggb_inst_src.c",path);
	}
	curr_out = fopen(src_fn,"w");
	if (!curr_out) {
		perror(src_fn);
		return NULL;
	}

	val = tr->u.fn_def.opt_expr;
	if (val->kind == op_raised) {
		compile_prol(val, &num_temps, &num_itemps);
		val = val->u.raised.expr;
	} else {
		compile_prol(NULL, &num_temps, &num_itemps);
	}
	compile_top_tree(val,&num_temps,&num_itemps,&done);
	compile_epil(done, hash, num_temps, num_itemps);

	fclose(curr_out);
	curr_out = NULL;

	if (num_errors) {
		do_warning("compiling failed");
		num_errors = save_num_errors;
		return NULL;
	}

	if (((dll_fn = run_compiler()) == NULL) ||
	    ((ret = load_dll(tr, dll_fn, hash)) == NULL)) {
		/* compilation failed */
		do_warning("compiler failed");
		num_errors = save_num_errors;
		return NULL;
	}

	cache_save(dll_fn, hash);
	unlink(dll_fn);

	num_errors = save_num_errors;
	return ret;
}


void
precompile(void)
{
	struct symbol *s;

	snd_enabled = 0;	/* because OS/X will run the sound thread */

	printf("\n");

	for (s = symtab; s; s = s->next) {
		if (!s->defn)
			continue;
		assert_str(s->defn->kind == op_fn_def, "fn defined with non-fn_def");
		if (!s->compiled) {
			/* XXX - this will compile nearly everything, even
			 * adsr...we should impose some limits, like only
			 * functions of exactly two args or something */
			printf("compiling ... %60s\r", s->name);
			fflush(stdout);
			s->compiled = compile_func(s->defn);
		}
	}
	printf("\n");

	snd_enabled = 1;
}


void
compiled_free(struct compiled_rec *bi)
{
	if (bi == NULL) return;
	bi->func = NULL;
	if (bi->dlh) {
		dlclose(bi->dlh);
		bi->dlh = NULL;
	}
	free(bi);
}

/* runs the compiler against src_fn, returning the filename for the
 * generated .so file */
static char *
run_compiler(void)
{
	static char *dll_fn = NULL;
	static int dll_uniq_start = 0;
	static int dll_counter = 0;
	char *argv[20];
	int argc;
	pid_t child;
#if defined(__APPLE_CC__)
	static char *obj_fn = NULL;
	static int obj_fn_len = 0;
#endif

	if (!dll_fn) {
		char *path = tmp_path();
		dll_fn = safe_malloc(strlen(path)+50);
		sprintf(dll_fn,"%s/ggb_dll_",path);
		dll_uniq_start = strlen(dll_fn);
	}
#if defined(__linux__)
	sprintf(dll_fn+dll_uniq_start, "%d.so", dll_counter++);

	argv[0] = getenv("CC");
	if (!argv[0]) argv[0] = "cc";
	argc = 1;
	argv[argc++] = "-shared";
#ifdef MAEMO
	argv[argc++] = "-mfloat-abi=softfp";
	argv[argc++] = "-mfpu=vfp";
#endif
#if 0
	argv[argc++] = "-g3";
#else
	argv[argc++] = "-O2";
#endif
	argv[argc++] = "-o";
	argv[argc++] = dll_fn;
	argv[argc++] = src_fn;
	argv[argc] = NULL;

	child = fork();
	if (child) {
		int status;
		if (waitpid(child,&status,0) <= 0) {
			do_error("waitpid failed: %s",strerror(errno));
			return NULL;
		}
		if (WEXITSTATUS(status)) {
			do_error("compile failed");
			return NULL;
		}
		return dll_fn;
	} else {
		exit(execvp(argv[0],argv));
	}
#elif defined(__APPLE_CC__)
	if (strlen(src_fn)+2 > obj_fn_len) {
		obj_fn_len = strlen(src_fn)+100;
		free(obj_fn);
		obj_fn = safe_malloc(obj_fn_len);
	}
	strcpy(obj_fn, src_fn);
	if (*obj_fn) obj_fn[strlen(obj_fn)-1] = 'o';

	sprintf(dll_fn+dll_uniq_start, "%d.dylib", dll_counter++);

	argv[0] = getenv("CC");
	if (!argv[0]) argv[0] = "cc";
	argc = 1;
	argv[argc++] = "-c";
	argv[argc++] = "-O2";
	argv[argc++] = "-o";
	argv[argc++] = obj_fn;
	argv[argc++] = src_fn;
//	argv[argc++] = "-g3";
	argv[argc] = NULL;

	child = fork();
	if (child) {
		int status;
		if (waitpid(child,&status,0) <= 0) {
			do_error("waitpid failed: %s",strerror(errno));
			return NULL;
		}
		if (WEXITSTATUS(status)) {
			do_error("compile failed");
			return NULL;
		}
	} else {
		exit(execvp(argv[0],argv));
	}

	argv[0] = getenv("LIBTOOL");
	if (!argv[0]) argv[0] = "libtool";
	argc = 1;
	argv[argc++] = "-dynamic";
	argv[argc++] = "-o";
	argv[argc++] = dll_fn;
	argv[argc++] = obj_fn;
	argv[argc++] = "-lgcc";
	argv[argc++] = "-lc";
	argv[argc] = NULL;

	child = fork();
	if (child) {
		int status;
		if (waitpid(child,&status,0) <= 0) {
			do_error("waitpid failed: %s",strerror(errno));
			return NULL;
		}
		if (WEXITSTATUS(status)) {
			do_error("libtool failed");
			return NULL;
		}
		return dll_fn;
	} else {
		exit(execvp(argv[0],argv));
	}
#endif

	/* should be impossible to reach here */
	return NULL;
}

typedef struct raw_waveform *wave_get_t(char *);
typedef void init_globals_t(wave_get_t *wg, value_t *st, builtin_func fg);

/* loads the .so file in file named by fn that corresponds to function tr.
 * check that the hash member of the object matches hash (if specified),
 * and allocate a builtin record on success. */
static struct compiled_rec *
load_dll(struct tree_rec *tr, char *fn, char *hash)
{
	struct compiled_rec *comp;
	char **shash;
	int *nt;
	init_globals_t *it;

	if (!fn || !tr)
		return NULL;

	assert_str(tr->kind == op_fn_def, "loading a dll for a non-function-definition");

	comp = safe_malloc(sizeof *comp);
	comp->num_args = tr->u.fn_def.num_formals;
	comp->dlh = dlopen(fn,RTLD_NOW);
	if (!comp->dlh) {
		do_error("dlopen(%s): %s",fn,dlerror());
		compiled_free(comp);
		return NULL;
	}

	if (!(shash = dlsym(comp->dlh, "hash")) ||
	    !*shash ||
	    (hash && strcmp(hash, *shash))) {
		do_error("dlsym(hash) does not match!");
		compiled_free(comp);
		return NULL;
	}

	comp->func = dlsym(comp->dlh,"compiled_func");
	if (!comp->func) {
		do_error("dlsym(compiled_func): %s",dlerror());
		compiled_free(comp);
		return NULL;
	}

	nt = dlsym(comp->dlh,"num_temps");
	if (nt) {
		comp->num_temps = *nt;
	} else {
		do_error("dlsym(num_temps): %s",dlerror());
		compiled_free(comp);
		return NULL;
	}
	nt = dlsym(comp->dlh,"num_itemps");
	if (nt) {
		comp->num_itemps = *nt;
	} else {
		do_error("dlsym(num_itemps): %s",dlerror());
		compiled_free(comp);
		return NULL;
	}

	it = dlsym(comp->dlh,"init_globals");
	if (it) {
		it(wave_get, sine_table, func_graph);
	} else {
		do_error("dlsym(init_globals): %s",dlerror());
		compiled_free(comp);
		return NULL;
	}

	return comp;
}


#define TAB_SPACE 100
static char *
tabs(int depth)
{
	static char tabs[TAB_SPACE];
	if (depth >= (TAB_SPACE/3)-1) {
		depth = (TAB_SPACE/3)-2;
	}
	memset(tabs,' ',depth*3);
	tabs[depth*3] = 0;
	return tabs;
}


/* spits out prologue context for the big expression */
/* r is the op_raised if we need to initialize them.. */
static void
compile_prol(struct tree_rec *r, int *num_temps, int *num_itemps)
{
	int i;

	fprintf(curr_out,"#include <stdlib.h>\n");
	fprintf(curr_out,"#include <stdio.h>\n");
	fprintf(curr_out,"#include <math.h>\n");
	fprintf(curr_out,"typedef double value_t;\n");
	fprintf(curr_out,"typedef struct { value_t pan, value; } svalue_t;\n");
	fprintf(curr_out,"typedef signed int intval_t;\n");
	fprintf(curr_out,"static value_t *sine_table;\n");
	fprintf(curr_out,"struct sound_rec { struct sound_rec *next; struct tree_rec *tr; unsigned long ofs; unsigned long note_up; int id; value_t *tmps; intval_t *itmps; int *buf; int done; };\n\n");

	fprintf(curr_out,"struct raw_waveform { struct raw_waveform *next; char *name; value_t *data; unsigned long len; };\n");
	fprintf(curr_out,"typedef struct raw_waveform *wave_get_t(char *);\n");
	fprintf(curr_out,"static wave_get_t *wave_get;\n\n");

	fprintf(curr_out,"typedef value_t (*builtin_func)(struct sound_rec*,value_t*,intval_t*,value_t,value_t,value_t,value_t,value_t,value_t,value_t,value_t,int *);\n");
	fprintf(curr_out,"static builtin_func func_graph;\n\n");

		/* __wave_func() is basically a clone of func_waveform().
		 * itmps[0] = done, tmps[0] = index, tmps[1] = last t */
	fprintf(curr_out,"static inline value_t __wave_func(value_t *tmps, intval_t *itmps, value_t t, char *fn) {\n");
	fprintf(curr_out,"\tstruct raw_waveform *wav = wave_get(fn);\n");
	fprintf(curr_out,"\tintval_t x;\n");
	fprintf(curr_out,"\tif (!wav) { itmps[0] = 1; return 0; }\n");
	fprintf(curr_out,"\tt = t * %s;\n",value_out(SEC_MULT/FREQ_MULT));
	fprintf(curr_out,"\tif (tmps[1] > t) { tmps[0] += (t + %s) - tmps[1]; }\n", value_out(SEC_MULT));
	fprintf(curr_out,"\telse { tmps[0] += t - tmps[1]; }\n");
	fprintf(curr_out,"\ttmps[1] = t;\n");
	fprintf(curr_out,"\tx = tmps[0];\n");
	fprintf(curr_out,"\tif ((x >= wav->len) || (tmps[0] < 0.0)) { itmps[0] = 1; return 0; }\n");
	fprintf(curr_out,"\treturn wav->data[(unsigned)x];\n");
	fprintf(curr_out,"}\n");

	fprintf(curr_out,"\n");

	fprintf(curr_out,"void init_globals(wave_get_t *wg, value_t *st, builtin_func fg) {\n");
	fprintf(curr_out,"\twave_get = wg;\n");
	fprintf(curr_out,"\tsine_table = st;\n");
	fprintf(curr_out,"\tfunc_graph = fg;\n");
	fprintf(curr_out,"}\n");

	fprintf(curr_out,"\n");

	fprintf(curr_out,"static inline double myfmod(double x, int y) {\n");
#if 0
	/* this one is slightly faster on some platforms in simple cases, but it
	 * is scary and a tiny gain (around 3%) - Greg 2009/11/30 */
	fprintf(curr_out,"\twhile (x < 0.0) { x += y; } while (x >= y) { x -= y; } return x;\n");
#else
	fprintf(curr_out,"\treturn ((x) - ((((int)(y))*((int)(((int)(x))/((int)(y)))))) + (((x) < 0) ? ((int)(y)) : 0));\n");
#endif
	fprintf(curr_out,"}\n");

	fprintf(curr_out,"\n");

	fprintf(curr_out,"int compiled_func(struct sound_rec *curr_sound, value_t *tmps, intval_t *itmps, value_t v1, value_t v2, value_t v3, value_t v4, value_t v5, value_t v6, value_t v7, value_t v8, int *buf, intval_t nents) {\n");	/* } */

	fprintf(curr_out,"%svalue_t t1",tabs(1));
	for (i = 1; i < MAX_LOCALS; i++) 
		fprintf(curr_out,",t%d",i+1);
	fprintf(curr_out,";\n");

	if (r && r->u.raised.num_commons) {
		fprintf(curr_out,"%svalue_t c0", tabs(1));
		for (i = 1; i < r->u.raised.num_commons; i++) {
			fprintf(curr_out,",c%d", i);
		}
		fprintf(curr_out,";\t/* common exprs */\n");
		fprintf(curr_out,"%sint c0_d", tabs(1));
		for (i = 1; i < r->u.raised.num_commons; i++) {
			fprintf(curr_out,",c%d_d", i);
		}
		fprintf(curr_out,";\t/* common exprs done */\n");
	}
	fprintf(curr_out,"%sintval_t i;\n", tabs(1));
	fprintf(curr_out,"%svalue_t r_pan, r_value;\n", tabs(1));

	if (r && r->u.raised.num_formals) {
		int guard = (*num_itemps)++;
		(*num_itemps)++;
		fprintf(curr_out,"\n/* RAISED FORMALS: */\n");
		fprintf(curr_out,"%sif (!itmps[%d]) {\n", tabs(1), guard);
		fprintf(curr_out,"%sitmps[%d] = 1;\n", tabs(2), guard);
		(*num_temps) += r->u.raised.num_formals;
		for (i = 0; i < r->u.raised.num_formals; i++) {
			char buf[40];
			char *done;	/* unused */
			sprintf(buf, "tmps[%d] = (", i);
			line_push(buf, 2);
			compile_tree(r->u.raised.formals[i],0,num_temps,num_itemps,&done,1);
			line_append(");");
			line_pop();
		}
		fprintf(curr_out,"%s}\n", tabs(1));
	}

	fprintf(curr_out,"%sfor (i = 0; i < nents; i++) {\n",tabs(1)); /*}*/

	if (r && r->u.raised.num_commons) {
		fprintf(curr_out,"\n/* COMMON EXPRESSIONS */\n");
		for (i = 0; i < r->u.raised.num_commons; i++) {
			char buf[40];
			char *done = NULL;
			sprintf(buf, "c%d = (", i);
			line_push(buf, 2);
			compile_tree(r->u.raised.commons[i],0,num_temps,num_itemps,&done,2);
			line_append(");");
			line_pop();
			fprintf(curr_out, "%sc%d_d = (%s);\n", tabs(1), i, (done ? done : "0/*nul*/"));
		}
	}

	fprintf(curr_out,"\n");
}


/* spits out epilogue context for the big expression */
static void
compile_epil(char *done, char *hash, int num_temps, int num_itemps)
{
	fprintf(curr_out,"%sr_value *= 32767.0; if (r_pan<0.0) { buf[i*2+0] += r_value; buf[i*2+1] += r_value*(1.0+r_pan); } else { buf[i*2+0] += r_value*(1.0-r_pan); buf[i*2+1] += r_value; }\n", tabs(2));
	fprintf(curr_out,"%scurr_sound->ofs++;\n",tabs(2));
	fprintf(curr_out,/*{*/"%s}\n",tabs(1));
	if (done) {
		fprintf(curr_out,"%sreturn (%s);\n",tabs(1),done);
	} else {
		fprintf(curr_out,"%sreturn (0);\n",tabs(1));
	}
	if (done && (done != done0) && (done != done1))
		free(done);
	fprintf(curr_out,/*{*/"}\n");

		/* export the hash information */
	fprintf(curr_out,"const char *hash = \"%s\";\n", hash);
		/* export the num_temps information */
	fprintf(curr_out,"const int num_temps = %d;\n", num_temps);
		/* export the num_itemps information */
	fprintf(curr_out,"const int num_itemps = %d;\n", num_itemps);
}

static char *
name_of_local(int local)
{
	static char buf[40];
	if (local >= MAX_LOCALS) {
		do_error("too many local intermediate values needed - increase MAX_LOCALS in compile.c!");
		return NULL;
	}
	sprintf(buf,"t%d",local+1);
	return buf;
}

#define LOCAL_FLAG_T_FREQ 0x8000000	/* tree is arg to WAV_FREQ function */

static void
compile_top_tree(struct tree_rec *tr, int *num_temps, int *num_itemps, char **done)
{
	if (tr->kind == op_stereo) {
		char *tdone = NULL;
		line_push("r_pan = ",2);
		compile_tree(tr->u.stereo.pan,0,num_temps,num_itemps,&tdone,2);
		line_append(";");
		line_pop();
		line_push("r_value = ",2);
		compile_tree(tr->u.stereo.value,0,num_temps,num_itemps,done,2);
		line_append(";");
		line_pop();
	} else {
		fprintf(curr_out,"%sr_pan = 0.0;\n", tabs(2));
		line_push("r_value = ",2);
		compile_tree(tr,0,num_temps,num_itemps,done,2);
		line_append(";");
		line_pop();
	}
}

/* Appends an expression for tr onto the current line, using the temporary
 * associated with local if necessary.  If persistent temps are
 * needed, num_temps or num_itemps is incremented.
 */
static void
compile_tree(struct tree_rec *tr, int local, int *num_temps, int *num_itemps, char **done, int depth)
{
	int flag_t_freq = local & LOCAL_FLAG_T_FREQ;
	int flag_t_freq_handled = 0;
	char *donel,*doner;

	*done = done0;	/* default for errors etc */

	local &= ~LOCAL_FLAG_T_FREQ;

	if (flag_t_freq) {
		line_append("(");
	}

	switch (tr->kind) {
		case op_error:
			line_append("err");
			do_error("op_error unacceptable in compiled code");
			return;
		case op_fn_def:
			line_append("err");
			do_error("op_fn_def unacceptable in compiled code");
		case op_fn_call: {
			struct tree_rec *fn_def = tr->u.fn_call.func->defn;

			if (fn_def) {
				line_append("err");
				do_error("func %s should have been inlined", tr->u.fn_call.func->name);
			} else if (!strcmp(tr->u.fn_call.func->name, "sin")) {
				if (tr->u.fn_call.num_args != 1) {
					line_append("err");
					do_error("sin undefined (wrong args)");
				} else {
					line_append("sine_table[(unsigned)");
					compile_tree(tr->u.fn_call.args[0],local|LOCAL_FLAG_T_FREQ,num_temps,num_itemps,done,depth);
					line_append("]");
				}
			} else if (!strcmp(tr->u.fn_call.func->name, "cos")) {
				if (tr->u.fn_call.num_args != 1) {
					line_append("err");
					do_error("cos undefined (wrong args)");
				} else {
					char buf[40];
					line_append("sine_table[(((intval_t)");
					compile_tree(tr->u.fn_call.args[0],local|LOCAL_FLAG_T_FREQ,num_temps,num_itemps,done,depth);
					sprintf(buf,")+%d)%%%d]", (int)(FREQ_MULT/4.0), (int)FREQ_MULT);
					line_append(buf);
				}
			} else if (!strcmp(tr->u.fn_call.func->name, "white")) {
				if (tr->u.fn_call.num_args != 0) {
					line_append("err");
					do_error("white undefined (wrong args)");
				} else {
					line_append("((((value_t)random()) / (((value_t)RAND_MAX)/2.0)) - 1.0)");
				}
			} else if (!strcmp(tr->u.fn_call.func->name, "exp")) {
				if (tr->u.fn_call.num_args != 1) {
					line_append("err");
					do_error("exp undefined (wrong args)");
				} else {
					line_append("exp(");
					compile_tree(tr->u.fn_call.args[0],local,num_temps,num_itemps,done,depth);
					line_append(")");
				}
			} else if (!strcmp(tr->u.fn_call.func->name, "graph")) {
				if (tr->u.fn_call.num_args != 2) {
					line_append("err");
					do_error("graph undefined (wrong args)");
				} else {
					line_append("func_graph(NULL,NULL,NULL,");
					compile_tree(tr->u.fn_call.args[0],local,num_temps,num_itemps,&donel,depth);
					line_append(",");
					compile_tree(tr->u.fn_call.args[1],local,num_temps,num_itemps,&doner,depth);
					line_append(",0,0,0,0,0,0,NULL)");
					*done = done_or(donel, doner);
				}
			} else if (!strcmp(tr->u.fn_call.func->name, "waveform")) {
				int tmp_start;
				int itmp_start;

				if (tr->u.fn_call.num_args != 2) {
					line_append("err");
					do_error("waveform undefined (wrong args)");
				} else {
					tmp_start = (*num_temps);
					(*num_temps) += 2;
					itmp_start = (*num_itemps);
					(*num_itemps) += 1;

					donel = safe_malloc(30);
					sprintf(donel,"(itmps[%d])",itmp_start);
					{
						char buf[50];
						sprintf(buf, "__wave_func(&tmps[%d],&itmps[%d],", tmp_start, itmp_start);
						line_append(buf);
					}
					compile_tree(tr->u.fn_call.args[0],local|LOCAL_FLAG_T_FREQ,num_temps,num_itemps,&doner,depth);
					*done = done_or(donel, doner);
					line_append(",");
					if (tr->u.fn_call.args[1]->kind == op_string) {
						line_append("\"");
						line_append(get_string_from_strtab(tr->u.fn_call.args[1]->u.string));
						line_append("\"");
					} else {
						line_append("err");
						do_error("waveform second arg should be string");
					}
					line_append(")");
				}
			} else {
				line_append("err");
				do_error("function %s undefined",tr->u.fn_call.func->name);
			}
		}	break;
		case op_add:
		case op_subtract:
			line_append("(");
			compile_tree(tr->u.operands[0],local+0,num_temps,num_itemps,&donel,depth);
			line_append((tr->kind == op_add) ? "+" : "-");
			compile_tree(tr->u.operands[1],local+1,num_temps,num_itemps,&doner,depth);
			line_append(")");
			if (tr->u.operands[0]->kind == op_number) {
				*done = doner;
			} else if (tr->u.operands[1]->kind == op_number) {
				*done = donel;
			} else {
				*done = done_and(donel,doner);
			}
			break;
		case op_multiply:
		case op_divide:
			line_append("(");
			compile_tree(tr->u.operands[0],local+0,num_temps,num_itemps,&donel,depth);
			line_append((tr->kind == op_multiply) ? "*" : "/");
			compile_tree(tr->u.operands[1],local+1,num_temps,num_itemps,&doner,depth);
			line_append(")");
			*done = done_or(donel,doner);
			break;
		case op_termseq: {
			char buf[80];
			int tmp_local = local;
			int itmp_for_idx = (*num_itemps)++;
			int itmp_for_countdown = (*num_itemps)++;
			int tmp_for_prev = (*num_temps)++;
			int itmp_for_zero_wait = (*num_itemps)++;
			static int label_num = 0;
			int i;
			int my_label_no = label_num++;

			if (!name_of_local(tmp_local)) {	/* couldn't get a temp for some reason */
				line_append("err");
				return;
			}

			/* the index variable will be set to -1 when we are
			 * done, so have that represent the "done state" */
			if (!tr->u.termseq.repeats) {
				*done = safe_malloc(30);
				sprintf(*done,"(itmps[%d] == -1)",itmp_for_idx);
			}

			/* note that itmp_for_idx is one greater than the
			 * corresponding entries in durations[] and exprs[]
			 * so that it defaults to the "-1" value */

			fprintf(curr_out,"%sif (itmps[%d] == -1) { %s = 0; goto ts_label_%d; }\n", tabs(depth), itmp_for_idx, name_of_local(tmp_local), my_label_no);
				/* if (countdown==0 && (zero_wait>max_wait || prev==0)) { */
				/* max_wait in this case is one tenth second */
			fprintf(curr_out,"%sif ((itmps[%d] == 0) && ((itmps[%d] > %s) || (tmps[%d] == 0.0))) {\n",tabs(depth),itmp_for_countdown,itmp_for_zero_wait,value_out(SEC_MULT/10.0), tmp_for_prev);
				/* if (idx >= num_intervals) done */
			fprintf(curr_out,"%sif (itmps[%d] >= %d) {\n",tabs(depth+1),itmp_for_idx,tr->u.termseq.num_intervals);
			fprintf(curr_out,"%sitmps[%d] = %d;\n",tabs(depth+2),itmp_for_idx, tr->u.termseq.repeats ? 0 : -1);
			fprintf(curr_out,"%s%s = 0;\n",tabs(depth+2),name_of_local(tmp_local));
			if (!tr->u.termseq.repeats) {
				fprintf(curr_out,"%sgoto ts_label_%d;\n",tabs(depth+2),my_label_no);
			}
			fprintf(curr_out,"%s}\n",tabs(depth+1));
			fprintf(curr_out,"%sitmps[%d]++;\n",tabs(depth+1),itmp_for_idx);
			fprintf(curr_out,"%sitmps[%d]=0;\n",tabs(depth+1),itmp_for_zero_wait);
			/* countdown = durations[i] */
			fprintf(curr_out,"%sswitch((unsigned)itmps[%d]) {\n",tabs(depth+1),itmp_for_idx);
			for (i = 0; i < tr->u.termseq.num_intervals; i++) {
				fprintf(curr_out,"%scase %d:\n",tabs(depth+2),i+1);
				if (!tr->u.termseq.zero_crossing[i]) {
					fprintf(curr_out,"%sitmps[%d] = %s;\n",tabs(depth+3),itmp_for_zero_wait,value_out(SEC_MULT));
				}
				sprintf(buf,"itmps[%d]=(intval_t)((",itmp_for_countdown);
				line_push(buf,depth+3);
				compile_tree(tr->u.termseq.durations[i], local, num_temps, num_itemps, &donel, depth+3);
				sprintf(buf,")*%s); break;", value_out(SEC_MULT));
				line_append(buf);
				line_pop();
			}
			fprintf(curr_out,"%s}\n",tabs(depth+1));
				/* tone lasts forever... */
			fprintf(curr_out,"%sif (itmps[%d] == 0) { itmps[%d] = -1; }\n",tabs(depth+1),itmp_for_countdown,itmp_for_countdown);
				/* lasts until note up... */
			fprintf(curr_out,"%selse if (itmps[%d] < 0) { itmps[%d] = -2; }\n",tabs(depth+1),itmp_for_countdown,itmp_for_countdown);
				/* if (countdown == 0) zero_wait++ */
			fprintf(curr_out,"%s} else if (itmps[%d] == 0) { itmps[%d]++; }\n",tabs(depth),itmp_for_countdown,itmp_for_zero_wait);
				/* countdown-- */
			fprintf(curr_out,"%sif (itmps[%d] > 0) { itmps[%d]--; }\n",tabs(depth),itmp_for_countdown,itmp_for_countdown);
				/* check for note up */
			fprintf(curr_out,"%selse if ((itmps[%d] == -2) && (curr_sound->ofs >= curr_sound->note_up)) { itmps[%d] = 0; }\n",tabs(depth),itmp_for_countdown,itmp_for_countdown);
				/* value = exprs[i] */
			fprintf(curr_out,"%sswitch((unsigned)itmps[%d]) {\n",tabs(depth),itmp_for_idx);
			for (i = 0; i < tr->u.termseq.num_intervals; i++) {
				fprintf(curr_out,"%scase %d:\n",tabs(depth+1),i+1);
				sprintf(buf,"%s=",name_of_local(tmp_local));
				line_push(buf,depth+2);
				compile_tree(tr->u.termseq.exprs[i], local, num_temps, num_itemps, &donel, depth+2);
				if (i+1 < tr->u.termseq.num_intervals) {
					line_append("; break;");
					line_pop();
				} else {
					/* the last termseq.. */
					line_append(";");
					line_pop();
				/* if (donel && index+1 == num_intervals) done=1 */
					fprintf(curr_out,"%sif (%s) itmps[%d] = -1;\n",tabs(depth+2), donel, itmp_for_idx);
					fprintf(curr_out,"%sbreak;\n",tabs(depth+2));
				}
			}
			fprintf(curr_out,"%sdefault: %s=0;\n",tabs(depth+1),name_of_local(tmp_local));
			fprintf(curr_out,"%s}\n",tabs(depth));
				/* if (countdown == 0) { */
			fprintf(curr_out,"%sif (itmps[%d] == 0) {\n",tabs(depth),itmp_for_countdown);
				/* initialize prev on first zero wait cycle */
			fprintf(curr_out,"%sif (itmps[%d] == 0) { tmps[%d] = %s; }\n",tabs(depth+1),itmp_for_zero_wait,tmp_for_prev,name_of_local(tmp_local));
				/* close enough to zero to fudge it */
			fprintf(curr_out,"%sif ((%s <= 0.03125) && (%s >= -0.03125)) { tmps[%d] = 0.0; }\n",tabs(depth+1),name_of_local(tmp_local),name_of_local(tmp_local),tmp_for_prev);
				/* crossing zero one way */
			fprintf(curr_out,"%selse if ((%s <= 0.0) && (tmps[%d] >= 0.0)) { tmps[%d] = 0.0; }\n",tabs(depth+1),name_of_local(tmp_local),tmp_for_prev,tmp_for_prev);
				/* crossing zero the other way */
			fprintf(curr_out,"%selse if ((%s >= 0.0) && (tmps[%d] <= 0.0)) { tmps[%d] = 0.0; }\n",tabs(depth+1),name_of_local(tmp_local),tmp_for_prev,tmp_for_prev);
				/* default to actually keeping prev recent */
			fprintf(curr_out,"%selse { tmps[%d] = %s; }\n",tabs(depth+1),tmp_for_prev,name_of_local(tmp_local));
			fprintf(curr_out,"%s}\n",tabs(depth));
			fprintf(curr_out,"ts_label_%d:\n",my_label_no);

				/* add the reference to the result in context */
			line_append(name_of_local(tmp_local));
		}	break;
		case op_number: {
			char buf[40];
			value_t num = tr->u.number;

			sprintf(buf,"%s",value_out(num));
			if (num == 0) *done = done1;
			line_append(buf);
		}	break;
		case op_argref: {		/* must be a top-level formal */
			char buf[40];
			if (tr->u.argref & ARG_RAISED_FORMAL) {
				int i = tr->u.argref & ~ARG_RAISED_FORMAL;
				sprintf(buf,"tmps[%d]", i);
			} else if (tr->u.argref & ARG_RAISED_COMMON) {
				int i = tr->u.argref & ~ARG_RAISED_COMMON;
				sprintf(buf,"c%d", i);
				{
					char *db = malloc(40);
					sprintf(db,"c%d_d", i);
					*done = db;
				}
			} else {
				assert_str(tr->u.argref < NUM_COMPILE_ARGS, "totally inappropriate op_argref in compile_tree()");
				sprintf(buf,"v%d",tr->u.argref+1);
			}
			line_append(buf);
		}	break;
		case op_string:
			assert_str(0, "strings unimplemented in compiler");
			break;
		case op_timest: {
			int tmpno = (*num_temps)++;
			char buf[40];
			if ((flag_t_freq & LOCAL_FLAG_T_FREQ) ||
			    tr->u.timest.kind == t_freq) {
				flag_t_freq_handled = 1;
				sprintf(buf,"(tmps[%d]=myfmod(((value_t)((tmps[%d]+((%s*",tmpno,tmpno,value_out(FREQ_MULT));
				line_append(buf);
				compile_tree(tr->u.timest.operand, local, num_temps, num_itemps, done, depth);
				sprintf(buf,")/%s)))),", value_out(SEC_MULT));
				line_append(buf);
				sprintf(buf,"%s))", value_out(FREQ_MULT));
				line_append(buf);
			} else {	/* t_default */
				sprintf(buf,"(tmps[%d]=(tmps[%d]+((",tmpno,tmpno);
				line_append(buf);
				compile_tree(tr->u.timest.operand, local, num_temps, num_itemps, done, depth);
				sprintf(buf,")/%s)))",value_out(SEC_MULT));
				line_append(buf);
			}
		}	break;
		case op_stereo:
			line_append("err");
			do_error("op_stereo unacceptable in compiled code");
			break;
		case op_one_tree:
			assert_str(0, "op_one_tree unacceptable in compiled code");
			break;
		case op_raised:
			assert_str(0, "op_raised unacceptable in compiled code");
			break;
	}

	if (flag_t_freq) {
		if (!flag_t_freq_handled) {
#if 0
			char buf[40];
			sprintf(buf,"%%%d",FREQ_MULT);
			line_append(buf);
#else
			assert_str(0,"flag_t_freq unhandled?");
#endif
		}
		line_append(")");
	}
}

#define LINE_SIZE_INCR 512

static struct line_rec {
	struct line_rec *next;
	char *str;
	size_t len;
	size_t size;
} *line_stack = NULL;

static void
line_push(char *s,int depth)
{
	struct line_rec *l = safe_malloc(sizeof *l);
	l->next = line_stack;
	l->str = safe_malloc(LINE_SIZE_INCR);
	l->len = 0;
	l->size = LINE_SIZE_INCR;
	line_stack = l;

	line_append(tabs(depth));
	line_append(s);
}

static void
line_append(char *s)
{
	int len,slen;
	assert_str((line_stack != NULL), "line_append() before line_push()?!");
	slen = strlen(s);
	len = line_stack->len + slen + 1;
	if (len >= line_stack->size) {
		char *nstr;
		len += LINE_SIZE_INCR;
		nstr = safe_malloc(len);
		memcpy(nstr, line_stack->str, line_stack->len+1);
		line_stack->str = nstr;
		line_stack->size = len;
	}
	memcpy(line_stack->str+line_stack->len, s, slen+1);
	line_stack->len += slen;
}

static void
line_pop(void)
{
	struct line_rec *l = line_stack;
	assert_str((l != NULL), "line_pop() before line_push()?!");
	line_stack = l->next;
	fprintf(curr_out,"%s\n",l->str);
	free(l->str);
	l->str = NULL;
	free(l);
}

static char *
done_and(char *d,char *e)
{
	char *ret;

	if (d == done0 || e == done0) {
		ret = done0;
	} else if (d == done1) {
		ret = e;
	} else if (e == done1) {
		ret = d;
	} else {
		ret = safe_malloc(strlen(d) + strlen(e) + 10);
		sprintf(ret,"(%s && %s)",d,e);
	}
	
	if ((d != done0) && (d != done1) && (d != ret)) free(d);
	if ((e != done0) && (e != done1) && (e != ret)) free(e);
	return ret;
}

static char *
done_or(char *d,char *e)
{
	char *ret;

	if (d == done1 || e == done1) {
		ret = done1;
	} else if (d == done0) {
		ret = e;
	} else if (e == done0) {
		ret = d;
	} else {
		ret = safe_malloc(strlen(d) + strlen(e) + 10);
		sprintf(ret,"(%s || %s)",d,e);
	}
	
	if ((d != done0) && (d != done1) && (d != ret)) free(d);
	if ((e != done0) && (e != done1) && (e != ret)) free(e);
	return ret;
}



/* helpers for cache_hash: */
static char *hash_buf = NULL;
static int hash_buf_sz = 0;
static int hash_buf_len = 0;

static void
hash_buf_add(char *fmt, ...)
{
	va_list ap;

	if (hash_buf_len + 1025 >= hash_buf_sz) {
		char *t;
		hash_buf_sz = hash_buf_len + 2048;
		t = safe_malloc(hash_buf_sz);
		memcpy(t, hash_buf, hash_buf_len);
		t[hash_buf_len] = 0;
		free(hash_buf);
		hash_buf = t;
	}

	va_start(ap, fmt);
	vsnprintf(hash_buf + hash_buf_len, 1024, fmt, ap);
	va_end(ap);

	hash_buf[hash_buf_len+1024] = 0;
	hash_buf_len = strlen(hash_buf);
}

static void
cache_hash_internal(struct tree_rec *tr)
{
	switch (tr->kind) {
		case op_error:
			break;

		case op_fn_def:
			hash_buf_add("%s[%d]", tr->u.fn_def.name->name, tr->u.fn_def.num_formals);
			cache_hash_internal(tr->u.fn_def.expr);
			break;

		case op_fn_call: {
			int i;

			if (!tr->u.fn_call.func || !tr->u.fn_call.func->name) {
				hash_buf_add("call(unk,");
			} else {
				hash_buf_add("call(%s,", tr->u.fn_call.func->name);
			}
			if (tr->u.fn_call.func && tr->u.fn_call.func->defn) {
				cache_hash_internal(tr->u.fn_call.func->defn);
			} else {
				hash_buf_add("internal");
			}
			for (i = 0; i < tr->u.fn_call.num_args; i++) {
				hash_buf_add(",");
				cache_hash_internal(tr->u.fn_call.args[i]);
			}
			hash_buf_add(")");
		}	break;

		case op_add:
		case op_subtract:
		case op_multiply:
		case op_divide:
			hash_buf_add("(");
			cache_hash_internal(tr->u.operands[0]);
			if (tr->kind == op_add) {
				hash_buf_add("+");
			} else if (tr->kind == op_subtract) {
				hash_buf_add("-");
			} else if (tr->kind == op_multiply) {
				hash_buf_add("*");
			} else {
				hash_buf_add("/");
			}
			cache_hash_internal(tr->u.operands[1]);
			hash_buf_add(")");
			break;

		case op_termseq: {
			int i;
			hash_buf_add("[");
			for(i = 0; i < tr->u.termseq.num_intervals; i++) {
				cache_hash_internal(tr->u.termseq.durations[i]);
				if (tr->u.termseq.zero_crossing &&
				    tr->u.termseq.zero_crossing[i]) {
					hash_buf_add("::");
				} else {
					hash_buf_add(":");
				}
				cache_hash_internal(tr->u.termseq.exprs[i]);
				hash_buf_add(",");
			}
			hash_buf_add(tr->u.termseq.repeats ? "]r" : "]");
		}	break;

		case op_number:
			hash_buf_add("%f",tr->u.number);
			break;

		case op_argref:
			hash_buf_add("arg%d",tr->u.argref);
			break;

		case op_string:
			hash_buf_add("str(%s)", get_string_from_strtab(tr->u.string));
			break;

		case op_timest:
			hash_buf_add("t%d*(",(int)tr->u.timest.kind);
			cache_hash_internal(tr->u.timest.operand);
			hash_buf_add(")");
			break;

		case op_stereo:
			hash_buf_add("{");
			cache_hash_internal(tr->u.stereo.pan);
			hash_buf_add(",");
			cache_hash_internal(tr->u.stereo.value);
			hash_buf_add("}");
			break;

		case op_one_tree:
			hash_buf_add("one_tree[%d] ", tr->u.one_tree.num_temps);
			cache_hash_internal(tr->u.one_tree.expr);
			break;
		case op_raised: {
			int i;
			hash_buf_add("(raised[");
			for (i = 0; i < tr->u.raised.num_formals; i++) {
				hash_buf_add("f%d:", i);
				if (i != 0) {
					hash_buf_add(",");
				}
				cache_hash_internal(tr->u.raised.formals[i]);
			}
			hash_buf_add(";");
			for (i = 0; i < tr->u.raised.num_commons; i++) {
				hash_buf_add("c%d:", i);
				if (i != 0) {
					hash_buf_add(",");
				}
				cache_hash_internal(tr->u.raised.commons[i]);
			}
			hash_buf_add("]:");
			cache_hash_internal(tr->u.raised.expr);
			hash_buf_add(")");
		}	break;
	}
}

/* generate a unique string for this tree_rec */
static char *
cache_hash(struct tree_rec *tr)
{
	hash_buf_len = 0;
	hash_buf_add("=>");
	cache_hash_internal(tr);

	return hash_buf;
}

/* returns a shorter version of the hash appropriate for filename */
static char *
cache_hash_filename(char *hash)
{
	static char *fn = NULL;
	static int fn_len = 0;
	unsigned int val = 3325;
	unsigned int val2 = 1087;	/* added to minimize collisions */

	assert_str(cache_dir != NULL, "cache_hash_filename with no cache_dir");

	if (strlen(cache_dir) + 40 >= fn_len) {
		fn_len = strlen(cache_dir) + 128;
		free(fn);
		fn = safe_malloc(fn_len);
	}

	for (; *hash; hash++) {
		val = (val * 37) + *hash;
		val2 = (val2 * 41) + *hash;
	}

#if defined(__linux__)
	sprintf(fn,"%s/ggb_inst_%08X%08X.so", cache_dir, val, val2);
#elif defined(__APPLE_CC__)
	sprintf(fn,"%s/ggb_inst_%08X%08X.dylib", cache_dir, val, val2);
#else
#error eek
#endif

	return fn;
}

/* search the cache directory for anything matching hash.
 * if found, load it and return a builtin handle.
 * tr is needed so load_dll knows where to link the builtin
 * (well really it just uses num_args) */
static struct compiled_rec *
cache_search(char *hash, struct tree_rec *tr)
{
	char *fn;
	struct stat sb;

	if (!cache_dir) return NULL;

	fn = cache_hash_filename(hash);
	if (stat(fn, &sb) == 0) {
		/* file exists, try it... */
		return load_dll(tr, fn, hash);
	}

	return NULL;
}

/* save a dll named "fn" into the cache directory using hash. */
static void
cache_save(char *fn, char *hash)
{
	char *ofn;

	if (!cache_dir) return;

	ofn = cache_hash_filename(hash);
	link(fn, ofn);
}

/* note that this returns a pointer to a static buffer that is overwritten
 * on each call. */
static char *
value_out(value_t val)
{
	static char buf[100];
	int i;
	sprintf(buf,"%.24f", val);
	for (i = strlen(buf)-1; (i > 2) && (buf[i-1] != '.') && (buf[i] == '0'); i--) {
		buf[i] = 0;
	}
	return buf;
}
