/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include "misc.h"
#include "snd.h"


#define TICKS_PER_BEAT 32
#define TICKS_SLOP 18		/* acceptable error for start of note */

#define KEY_WIDTH 11
#define KEY_HEIGHT 80
#define KEYS 88			/* number of keys in our keyboard */
#define KEY_START 21		/* number of low keys to skip */
#define TICK_HEIGHT 5
#define HEIGHT 600
#define PREVIEW_HEIGHT (HEIGHT-KEY_HEIGHT)
#define WIDTH (KEY_WIDTH*KEYS)

struct gyro_note_rec {
	struct gyro_note_rec *next;
	int start;	/* absolute, in ticks */
	int stop;	/* absolute, in ticks */
	int dur;	/* specified by the note value, in ticks */
	int len;	/* 1 = staccato ... 4 = legato */
	int note;
	int vel;
	int finger;	/* fingering advice...left=4-0, right=5-9 */
};

static struct gyro_note_rec *notes, *curr_notes;
static int curr_tick;
static int curr_countdown;
static int samps_per_tick;
static int ticks_per_meas;
static int gyro_mode;
static int notes_seen;
static int notes_correct;		/* keeping score */

static Display *disp;
static int screen_num;
static Window root_win;
static GC gc;
static XColor col_white, col_dim_white;
static XColor col_black, col_dim_black;
static XColor col_left, col_right;
static XColor col_divide;
static XColor col_text;
static XColor col_bar1, col_bar2;
static Window win;


static int gyro_window(void);
static void gyro_redraw(void);
static void gyro_reset(void);
static int gyro_load(char *fn);

/* mode=1 for gyro playing, mode=2 for you playing */
void
gyro_start(char *fn, int bpm, int mode, int ofs)
{
	gyro_mode = mode;

	gyro_reset();
	if (!gyro_window()) {
		gyro_reset();
		return;
	}
	if (!gyro_load(fn)) {
		printf("failed to load %s\n",fn);
		return;
	}

	if (ofs) {
		for (curr_notes = notes;
		     curr_notes && (curr_notes->start < (ofs*ticks_per_meas));
		     curr_notes = curr_notes->next) {
		}
		curr_tick = curr_notes->start-1;
	} else {
		/* give some silence at the beginning */
		curr_tick = -ticks_per_meas;

		/* start at beginning of note list */
		curr_notes = notes;
	}

	if (bpm != -1) {
		samps_per_tick = ((44100*60 + (bpm / 2)) / bpm) / TICKS_PER_BEAT;
	}
}

void
gyro_stop(void)
{
	gyro_reset();
	if (gyro_x_fd != -1) {
		gyro_redraw();
	}
}

void
gyro_step(void)
{
	struct gyro_note_rec *n;
	static int redraw_counter = 0;

	if (!curr_notes) {
		if (notes) {
			/* end of song */
			gyro_reset();
		}
		return;
	}

	if (++redraw_counter == 2205) {
		redraw_counter = 0;
		gyro_redraw();
	}

	curr_countdown--;
	if (curr_countdown) {
		return;
	}

	curr_countdown = samps_per_tick;
	curr_tick++;

	n = curr_notes;
	while (n) {
		while ((n->note < 0) && (n->note > 127)) {
			n = n->next;
		}
		if (gyro_mode&1) {
			if (n->stop == curr_tick) {
				cmd_note(n->note, 0);
			}
			if (n->start == curr_tick) {
				cmd_note(n->note, n->vel);
			}
		}
		if (n->start == curr_tick) {
			notes_seen++;
		}
		if (n->start > curr_tick) {
			break;
		}
		n = n->next;
	}

	/* now skip past any totally done notes */
	while (curr_notes) {
		if (curr_notes->stop <= curr_tick) {
			curr_notes = curr_notes->next;
		} else {
			break;
		}
	}
}

int
gyro_note(unsigned char note, unsigned char vel)
{
	struct gyro_note_rec *n;
	if (!(gyro_mode&2)) {
		return 0;
	}
	if (!vel) {
		/* only keep score of note downs */
		return 0;
	}
	for (n = curr_notes; n; n = n->next) {
		if ((n->note == note) &&
		    (n->start >= curr_tick-TICKS_SLOP) &&
		    (n->start <= curr_tick+TICKS_SLOP)) {
			notes_correct++;
			return 0;
		}
		if (n->start > curr_tick+TICKS_SLOP) {
			break;
		}
	}
	return 1;		/* inappropriate note */
}

void
gyro_poll(void)
{
	XEvent evt;

	while (XPending(disp)) {
		XNextEvent(disp, &evt);
		/* we don't really respond to events so whatever */
	}
}

void gyro_cleanup(void)
{
	gyro_reset();

	XCloseDisplay(disp);
	gyro_x_fd = -1;
}


static int
gyro_window(void)
{
	Colormap cm;

	if (gyro_x_fd != -1) {
		return 1;
	}

	if (!(disp = XOpenDisplay(getenv("DISPLAY")))) {
		fprintf(stderr,"Failed to connect to X server.  Is DISPLAY right?\n");
		return 0;
	}

	screen_num = DefaultScreen(disp);
	root_win = RootWindow(disp, screen_num);

	gc = XCreateGC(disp, root_win, 0, 0);
	XSetFillStyle(disp, gc, FillSolid);
	cm = DefaultColormap(disp, screen_num);
	col_white.red   = 0xFFFF;
	col_white.green = 0xFFFF;
	col_white.blue  = 0xFFFF;
	XAllocColor(disp, cm, &col_white);
	col_dim_white.red   = 0x6666;
	col_dim_white.green = 0x6666;
	col_dim_white.blue  = 0x6666;
	XAllocColor(disp, cm, &col_dim_white);
	col_black.red   = 0x0;
	col_black.green = 0x0;
	col_black.blue  = 0x0;
	XAllocColor(disp, cm, &col_black);
	col_dim_black.red   = 0x0;
	col_dim_black.green = 0x0;
	col_dim_black.blue  = 0x0;
	XAllocColor(disp, cm, &col_dim_black);
	col_left.red   = 0xFFFF;
	col_left.green = 0x3333;
	col_left.blue  = 0x3333;
	XAllocColor(disp, cm, &col_left);
	col_right.red   = 0x3333;
	col_right.green = 0xFFFF;
	col_right.blue  = 0x3333;
	XAllocColor(disp, cm, &col_right);
	col_divide.red   = 0xAAAA;
	col_divide.green = 0xAAAA;
	col_divide.blue  = 0xAAAA;
	XAllocColor(disp, cm, &col_divide);
	col_text.red   = 0x0;
	col_text.green = 0x0;
	col_text.blue  = 0xFFFF;
	XAllocColor(disp, cm, &col_text);
	col_bar1.red   = 0x8888;
	col_bar1.green = 0x0;
	col_bar1.blue  = 0x8888;
	XAllocColor(disp, cm, &col_bar1);
	col_bar2.red   = 0x0;
	col_bar2.green = 0xFFFF;
	col_bar2.blue  = 0xFFFF;
	XAllocColor(disp, cm, &col_bar2);

	win = XCreateSimpleWindow(disp,root_win,0,0,
		WIDTH, HEIGHT, 0,
		BlackPixel(disp,screen_num),
		WhitePixel(disp,screen_num));
	XStoreName(disp, win, "ggb piano gyro");
	/* we have to accept these, right? */
	XSelectInput(disp, win, ExposureMask);
	XMapWindow(disp, win);

	XFlush(disp);

	gyro_x_fd = ConnectionNumber(disp);

	return 1;
}

static char *
finger_name(int finger)
{
	static char ret[2];
	ret[1] = 0;
	if (finger < 5) {
		ret[0] = '5'-finger;
	} else {
		ret[0] = '1'+(finger-5);
	}
	return ret;
}

static void
draw_key(int note, int finger, int when, int dur)
{
	enum { white_left, white_middle, white_right, black } kind = white_left;
	int wkw = ((KEY_WIDTH*12)+3)/7;
	int ky = HEIGHT-KEY_HEIGHT;
	int kh = KEY_HEIGHT;
	int hkh = KEY_HEIGHT/2;
	int bottom_left;
	int x;

		/* octave offset: */
	bottom_left = x = (note-KEY_START)*KEY_WIDTH;
	note %= 12;

	switch (note) {
		case 0: kind = white_left;	break;
		case 1: kind = black;		break;
		case 2: kind = white_middle;
			bottom_left -= (2*KEY_WIDTH)/7;
			break;
		case 3: kind = black;		break;
		case 4: kind = white_right;
			bottom_left -= (4*KEY_WIDTH)/7;
			break;
		case 5: kind = white_left;	break;
		case 6: kind = black;		break;
		case 7: kind = white_middle;
			bottom_left -= (1*KEY_WIDTH)/7;
			break;
		case 8: kind = black;		break;
		case 9: kind = white_middle;
			bottom_left -= (3*KEY_WIDTH)/7;
			break;
		case 10: kind = black;		break;
		case 11: kind = white_right;
			bottom_left -= (5*KEY_WIDTH)/7;
			break;
	}

	if ((finger >= 0) && (finger < 5)) {
		XSetForeground(disp, gc, col_left.pixel);
	} else if (finger >= 5) {
		XSetForeground(disp, gc, col_right.pixel);
	} else if (kind == black) {
		XSetForeground(disp, gc, col_black.pixel);
	} else {
		XSetForeground(disp, gc, col_white.pixel);
	}
	if (when <= 0) {
		switch (kind) {
			case black:
				XFillRectangle(disp, win, gc, x, ky, KEY_WIDTH, hkh);
				break;
			case white_left:
				XFillRectangle(disp, win, gc, x, ky, KEY_WIDTH, hkh);
				XFillRectangle(disp, win, gc, x, ky+hkh, wkw, hkh);
				XSetForeground(disp, gc, col_divide.pixel);
				XDrawLine(disp, win, gc, x, ky, x, HEIGHT);
				break;
			case white_middle:
			case white_right:
				XFillRectangle(disp, win, gc, x, ky, KEY_WIDTH, hkh);
				XFillRectangle(disp, win, gc, bottom_left, ky+hkh, wkw, hkh);
				XSetForeground(disp, gc, col_divide.pixel);
				XDrawLine(disp, win, gc, bottom_left, ky+hkh, bottom_left, HEIGHT);
				break;
		}
		if (finger >= 0) {
			XSetForeground(disp, gc, col_text.pixel);
			if (kind == black) {
				XDrawString(disp, win, gc, x, ky+hkh, finger_name(finger), 1);
			} else {
				XDrawString(disp, win, gc, x, ky+kh, finger_name(finger), 1);
			}
		}
	}
	if (finger == -1) {
		if (kind == black) {
			XSetForeground(disp, gc, col_dim_black.pixel);
		} else {
			XSetForeground(disp, gc, col_dim_white.pixel);
		}
		XFillRectangle(disp, win, gc, x, 0, KEY_WIDTH, ky);
		if (kind == white_left) {
			XSetForeground(disp, gc, col_divide.pixel);
			XDrawLine(disp, win, gc, x, 0, x, ky);
		}
	}
	if ((when+dur) > 0) {
		int bottom, top;
		when *= TICK_HEIGHT;
		dur *= TICK_HEIGHT;
		bottom = ky-((when<0)?0:when);
		top = ky-(when+dur);
		if (top < 0) {
			top = 0;
		}
		if (finger < 5) {
			XSetForeground(disp, gc, col_left.pixel);
		} else {
			XSetForeground(disp, gc, col_right.pixel);
		}
		XFillRectangle(disp, win, gc, x, top, KEY_WIDTH, bottom-top);
		XSetForeground(disp, gc, col_text.pixel);
		XDrawString(disp, win, gc, x, bottom, finger_name(finger), 1);
	}
}

static void
gyro_redraw(void)
{
	struct gyro_note_rec *n;
	char buf[80];
	int i;

	/* first, draw keyboard and background */
	for (i = KEY_START; i < KEY_START+KEYS; i++) {
		draw_key(i, -1, -1, -1);
	}

	for (n = notes; n; n = n->next) {
		if (n->stop < curr_tick) {
			continue;
		}
		if (n->start > curr_tick + (PREVIEW_HEIGHT/TICK_HEIGHT)) {
			break;
		}
		if ((n->note <= KEY_START) || (n->note > (KEY_START+KEYS))) {
			continue;
		}
		draw_key(n->note, n->finger, n->start-curr_tick, n->stop-n->start);
	}

	XSetForeground(disp, gc, col_bar2.pixel);
	XDrawLine(disp, win, gc, 0, HEIGHT-KEY_HEIGHT, WIDTH, HEIGHT-KEY_HEIGHT);
	XSetForeground(disp, gc, col_bar1.pixel);
	i = curr_tick;
	while (i < 0) {
		i += ticks_per_meas;
	}
	for (i = (HEIGHT-KEY_HEIGHT) -
		 (TICK_HEIGHT*(ticks_per_meas-(i%ticks_per_meas)));
	     i >= 0;
	     i -= TICK_HEIGHT*ticks_per_meas) {
		XDrawLine(disp, win, gc, 0, i, WIDTH, i);
	}

	XSetForeground(disp, gc, col_text.pixel);
	sprintf(buf,"notes seen: %d", notes_seen);
	XDrawString(disp, win, gc, 0, 20, buf, strlen(buf));
	sprintf(buf,"correct: %d (%.1f%%)", notes_correct, (((double)notes_correct)/notes_seen)*100.0);
	XDrawString(disp, win, gc, 0, 40, buf, strlen(buf));

	XFlush(disp);
}

static void
gyro_reset(void)
{
	struct gyro_note_rec *n;
	while (notes) {
		n = notes;
		notes = notes->next;
		free(n);
	}
	curr_notes = NULL;
	curr_tick = 0;
	samps_per_tick = 22050;
	curr_countdown = 1;
	ticks_per_meas = 4*TICKS_PER_BEAT;

	notes_seen = notes_correct = 0;

	snd_shutup();
}

static int
read_int(FILE *f, int *cp)
{
	int c;
	int ret = 0;
	while (1) {
		c = fgetc(f);
		if (isdigit(c)) {
			ret = (ret*10) + (c-'0');
		} else {
			break;
		}
	}
	*cp = c;
	return ret;
}

/* called with curr==-1 to say that this call is just to reset the
 * fingering table */
static int
compute_next_finger(int prev, int curr, int prev_finger)
{
	static int last_fingers[10];	/* notes played by last fingers */
	int i;
	int hand;

	last_fingers[prev_finger] = prev;

	hand = ((prev_finger < 5) ? 0 : 5);
	prev_finger -= hand;

	if (curr == -1) {
		/* reset table */
		for (i = 0; i < 5; i++) {
			if (i != prev_finger) {
				last_fingers[i+hand] = -1;
			}
		}
		return -1;
	}

	/* first, if we have recently hit this note with the same hand, use
	 * the same finger */
	for (i = 0; i < 5; i++) {
		if (last_fingers[i+hand] == curr) {
			return i+hand;
		}
	}

	if (curr < prev) {
		i = prev_finger;
		while ((curr < prev) && (i > 0)) {
			curr += 3;
			i--;
		}
	} else {
		i = prev_finger;
		while ((curr > prev) && (i < 4)) {
			curr -= 3;
			i++;
		}
	}
	i += hand;
	last_fingers[i] = curr;

	return i;
}

static int
gyro_load(char *fn)
{
	FILE *f;
	struct gyro_note_rec *curr, *last = NULL;
	int curr_done;
	int c;

	f = fopen(fn, "r");
	if (!f) {
		perror(fn);
		return 0;
	}

	curr = safe_malloc(sizeof *curr);
	memset(curr, 0, sizeof *curr);
	curr->dur = TICKS_PER_BEAT;
	curr->len = 3;
	curr->note = 50;
	curr->vel = 127;
	curr->finger = -1;
	curr_done = 1;

	do {
		c = fgetc(f);
retry: ;
		if (((c == EOF) || isspace(c) || (c == '%')) &&
		    !curr_done) {
			int len;
			struct gyro_note_rec *prev;
			prev = last;
			if (!prev || (prev->start > curr->start)) {
				prev = notes;
			}
			while (prev && prev->next &&
			       (prev->next->start < curr->start)) {
				prev = prev->next;
			}
			if (prev) {
				if (prev->start > curr->start) {
					curr->next = notes;
					notes = curr;
				} else {
					curr->next = prev->next;
					prev->next = curr;
				}
			} else {
				notes = curr;
			}
			if (curr->note != -1) {
				if (curr->finger == -1) {
					if (last && last->note != -1) {
						curr->finger = compute_next_finger(last->note, curr->note, last->finger);
					} else {
						curr->finger = 0;
					}
				} else {
					(void)compute_next_finger(curr->note, -1, curr->finger);
				}
			}

			len = (curr->dur * curr->len) / 4;
			if (len < 1) {
				len = 1;
			}
			curr->stop = curr->start + len;
			last = curr;
			curr = safe_malloc(sizeof *curr);
			*curr = *last;
			curr->next = NULL;
			curr->start = last->start + last->dur;
			curr->finger = -1;
			curr_done = 1;
		}
		switch (c) {
			case '%':
				do {
					c = fgetc(f);
				} while ((c != EOF) && (c != '\n'));
				break;
			case 'C':
				curr_done = 0;
				curr->note = ((curr->note/12)*12) + 0;
				break;
			case 'D':
				curr_done = 0;
				curr->note = ((curr->note/12)*12) + 2;
				break;
			case 'E':
				curr_done = 0;
				curr->note = ((curr->note/12)*12) + 4;
				break;
			case 'F':
				curr_done = 0;
				curr->note = ((curr->note/12)*12) + 5;
				break;
			case 'G':
				curr_done = 0;
				curr->note = ((curr->note/12)*12) + 7;
				break;
			case 'A':
				curr_done = 0;
				curr->note = ((curr->note/12)*12) + 9;
				break;
			case 'B':
				curr_done = 0;
				curr->note = ((curr->note/12)*12) + 11;
				break;
			case 'n':
				curr_done = 0;
				curr->note = 0;
				break;
			case 's':
				curr->start = read_int(f,&c) * ticks_per_meas;
				goto retry;
			case '<':
				if (last) {
					curr->start = last->start;
				}
				break;
			case 'R': {
				int bpm;
				bpm = read_int(f,&c);	
				samps_per_tick = ((44100*60 + (bpm / 2)) / bpm) / TICKS_PER_BEAT;
				goto retry;
			}
			case 'M':
				ticks_per_meas = read_int(f,&c) * (TICKS_PER_BEAT/2);
				goto retry;
			case 'v':
				curr_done = 0;
				curr->vel = read_int(f,&c);
				goto retry;
			case '*':
				curr_done = 0;
				curr->len = 1;
				break;
			case '^':
				curr_done = 0;
				curr->len = 2;
				break;
			case '-':
				curr_done = 0;
				curr->len = 3;
				break;
			case '_':
				curr_done = 0;
				curr->len = 4;
				break;
			case 'l':
				curr_done = 0;
				curr->finger = 5-read_int(f,&c);
				compute_next_finger(-1, -1, -1);
				/* all fingering bets are off */
				goto retry;
			case 'r':
				curr_done = 0;
				curr->finger = 4+read_int(f,&c);
				goto retry;
			case 'd':
				curr_done = 0;
				curr->dur = (4*TICKS_PER_BEAT)/read_int(f,&c);
				goto retry;
			case 'o':
				curr_done = 0;
				c = fgetc(f);
				if (c == '+') {
					curr->note += 12;
				} else if (c == '-') {
					curr->note -= 12;
				} else if (isdigit(c)) {
					curr->note = (curr->note%12) + ((c-'0')*12);
				} else if (c == 'A') {
					curr->note = (curr->note%12) + 120;
				}
				break;
			case '#':
				curr_done = 0;
				curr->note++;
				break;
			case 'b':
				curr_done = 0;
				curr->note--;
				break;
			default:
				if (!isspace(c) && (c != EOF)) {
					printf("ignored invalid character \"%c\"\n",c);
				}
				break;
		}
	} while (c != EOF);

	fclose(f);

	return 1;
}
