/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "misc.h"


struct help_rec {
	struct help_rec *next;
	struct help_rec *parent;
	struct help_rec *children;
	char *full_title;
	char *title;
	char *summary;
	char *detail;
};

static struct help_rec *help_root = NULL;


static struct help_rec *find_help(char *s, struct help_rec *curr);
static void display_help(struct help_rec *hr);
static void read_help(void);
static void help_add(char *title, char *summary, char *detail);
static struct help_rec *help_make_title(struct help_rec *curr, char *title);


void
help(char *args)
{
	static struct help_rec *help_last = NULL;
	struct help_rec *curr = NULL;

	if (!help_root)
		read_help();

	args += strspn(args," \t\r\n");

	if (*args) {
		struct help_rec *p = (help_last ? help_last : help_root);
		do {
			if (!p)
				break;
			curr = find_help(args, p);
			p = p->parent;
		} while (!curr);
	} else {
		curr = help_root;
	}
	if (!curr) {
		printf("help for \"%s\" not found.\n",args);
		return;
	}

	if (curr->children) {
		help_last = curr;
	}

	display_help(curr);
}

static struct help_rec *
find_help(char *s, struct help_rec *curr)
{
	while (*s && curr) {
		int len = strcspn(s," \t\r\n");
		struct help_rec *child;
		struct help_rec *candidate = NULL;

		for (child = curr->children; child; child = child->next) {
			if (child->title &&
			    !strncmp(child->title, s, len)) {
				if (strlen(child->title) == len) {
					candidate = child;
					break;
				} else if (!candidate) {
					candidate = child;
				} else {
					candidate = (void *)-1;
				}
			}
		}
		if (candidate == (void *)-1) {
			curr = NULL;
			printf("\"%s\" is ambiguous.\n", s);
		} else {
			curr = candidate;
		}

		s += len;
		s += strspn(s," \t");
	}

	return curr;
}

static void
display_help(struct help_rec *hr)
{
	struct help_rec *p;
	int max_len = 0;

	printf("help for %s:\n",(hr->full_title && *hr->full_title) ? hr->full_title : "ggb");
	if (hr->detail) {
		printf("%s\n", hr->detail);
	} else if (hr->summary) {
		printf("%s\n", hr->summary);
	}

	if (!hr->children)
		return;

	for (p = hr->children; p; p = p->next) {
		if (p->title && strlen(p->title) > max_len)
			max_len = strlen(p->title);
	}

	printf("\nsub-topics:\n");
	for (p = hr->children; p; p = p->next) {
		if (p->title && p->summary) {
			printf("  %-*s  %s\n",max_len, p->title, p->summary);
		}
	}
}

static void
read_help(void)
{
	FILE *f = open_search(HELP_FILE,"r");
	char buf[HELP_MAX_LINE];

	if (!f) {
		perror(HELP_FILE);
		return;
	}

	while (fgets(buf,sizeof buf,f)) {
		char *title = NULL, *summary = NULL, *detail = NULL;
		buf[strcspn(buf,"\r\n")] = 0;
		title = buf;
		summary = strchr(title,'|');
		if (summary) {
			*summary = 0;
			summary++;
			detail = strchr(summary,'|');
			if (detail) {
				*detail = 0;
				detail++;
			}
		}
		if (title && !*title) summary = NULL;
		if (summary && !*summary) summary = NULL;
		if (detail && !*detail) detail = NULL;
		if (title || summary || detail) {
			help_add(title, summary, detail);
		}
	}

	fclose(f);
}

static void
help_add(char *title, char *summary, char *detail)
{
	struct help_rec *curr;
	char *full_title = title ? safe_strdup(title) : NULL;

	if (!help_root) {
		help_root = safe_malloc(sizeof *help_root);
		memset(help_root,0,sizeof *help_root);
	}
	curr = help_root;

	while (title && *title) {
		int len = strcspn(title," \t\r\n");
		if (title[len]) {
			title[len] = 0;
			len++;
		}

		curr = help_make_title(curr,title);

		title += len;
	}

	curr->full_title = full_title;

	if (summary) curr->summary = safe_strdup(summary);
	if (detail) curr->detail = safe_strdup(detail);
}

static struct help_rec *
help_make_title(struct help_rec *curr, char *title)
{
	struct help_rec **p;
	struct help_rec *n;

	for (p = &(curr->children); *p; p = &((*p)->next)) {
		if ((*p)->title && !strcmp((*p)->title, title))
			return *p;
	}

	n = safe_malloc(sizeof *n);
	memset(n, 0, sizeof *n);
	n->title = safe_strdup(title);
	n->parent = curr;

	*p = n;

	return n;
}
