/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: lexer.c,v 1.11 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "lang.h"
#include "midi.h"

#define MAX_LEN 1024		/* the longest single token */

static int curr_ch = 0;

#define DO_TOK(x) { curr_token = tok_##x; state = state_done; }

void
lexer_startup(void)
{
	curr_ch = lex_getchar();
}

token
next_token(void)
{
	enum {
		state_start,
		state_colon,
		state_sym,
		state_num,
		state_num_exp1,
		state_num_exp2,
		state_str,
		state_str_escape,
		state_comment,
		state_note,
		state_done
	} state = state_start;
	static char buf[MAX_LEN+1];
	unsigned int buf_pos = 0;

	while (1) {
retry:
		switch(state) {
		case state_start:
			switch(curr_ch) {
			case 0: case ' ': case '\t': case '\n': case '\r':
				break; /* ignore whitespace */
			case ';': DO_TOK(semicolon) break;
			case '(': DO_TOK(lparen) break;
			case ')': DO_TOK(rparen) break;
			case '[': DO_TOK(lbracket) break;
			case ']': DO_TOK(rbracket) break;
			case '{': DO_TOK(lbrace) break;
			case '}': DO_TOK(rbrace) break;
			case ',': DO_TOK(comma) break;
			case ':': state = state_colon; break;
			case '=': DO_TOK(equals) break;
			case '+': DO_TOK(plus) break;
			case '-': DO_TOK(minus) break;
			case '*': DO_TOK(star) break;
			case '/': DO_TOK(slash) break;
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
			case '.': state = state_num; goto retry;
			case '"': state = state_str; break;
			case '\'': state = state_note; break;
			case '#': state = state_comment; break;
			case EOF: curr_token = tok_eof; return curr_token;
			default:
				if (isalpha(curr_ch) || curr_ch == '_') {
					state = state_sym; goto retry;
				} else {
					do_error("unknown token");
					DO_TOK(error) break;
				}
			}
			break;
		case state_colon:
			if (curr_ch == ':') {
				DO_TOK(coloncolon) break;
			} else {
				curr_token = tok_colon;
				return curr_token;
			}
		case state_sym:
			if ((isalnum(curr_ch) || curr_ch == '_') &&
			    buf_pos < MAX_LEN) {
				buf[buf_pos++] = curr_ch;
				break;
			} else {
				buf[buf_pos] = 0;
				curr_token = tok_symbol;
				curr_token_sym = sym_lookup(buf);
				if (buf_pos >= MAX_LEN) {
					do_warning("symbol name too long: %s",buf);
				}
				return curr_token;
			}
		case state_num:
			if ((isdigit(curr_ch) || curr_ch == '.') &&
			    buf_pos < MAX_LEN) {
				buf[buf_pos++] = curr_ch;
				break;
			} else if (curr_ch == 'e' && buf_pos < MAX_LEN) {
				buf[buf_pos++] = curr_ch;
				state = state_num_exp1;
				break;
			} else {
				buf[buf_pos] = 0;
				curr_token = tok_number;
				curr_token_num = atof(buf);
				if (buf_pos >= MAX_LEN) {
					do_warning("number too long: %s",buf);
				}
				return curr_token;
			}
		case state_num_exp1:	/* accept the sign of the exponent */
			if ((isdigit(curr_ch) ||
			     curr_ch == '+' || curr_ch == '-') &&
			    buf_pos < MAX_LEN) {
				buf[buf_pos++] = curr_ch;
				state = state_num_exp2;
				break;
			} else {
				buf[buf_pos] = 0;
				curr_token = tok_number;
				curr_token_num = atof(buf);
				if (buf_pos >= MAX_LEN) {
					do_warning("number too long: %s",buf);
				}
				return curr_token;
			}
		case state_num_exp2:
			if (isdigit(curr_ch) && buf_pos < MAX_LEN) {
				buf[buf_pos++] = curr_ch;
				break;
			} else {
				buf[buf_pos] = 0;
				curr_token = tok_number;
				curr_token_num = atof(buf);
				if (buf_pos >= MAX_LEN) {
					do_warning("number too long: %s",buf);
				}
				return curr_token;
			}
		case state_str:
			if (curr_ch == EOF) {
				do_warning("unexpected EOF in string");
				curr_token = tok_eof;
				return curr_token;
			} else if (curr_ch == '"' || buf_pos >= MAX_LEN) {
				buf[buf_pos] = 0;
				curr_token_str = buf;
				curr_token = tok_string;
				state = state_done;
				if (buf_pos >= MAX_LEN) {
					do_warning("string too long: %s",buf);
				}
				break;
			} else if (curr_ch == '\\') {
				state = state_str_escape;
				break;
			} else {
				buf[buf_pos++] = curr_ch;
				break;
			}
		case state_str_escape:
			/* we always have space for at least one more,
			 * or we wouldn't be in this state. */
			switch (curr_ch) {
			case EOF: curr_token = tok_eof; return curr_token;
			case 'r': buf[buf_pos++] = '\r'; break;
			case 'n': buf[buf_pos++] = '\n'; break;
			case 't': buf[buf_pos++] = '\t'; break;
			default: buf[buf_pos++] = curr_ch; break;
			}
			state = state_str;
			break;
		case state_comment:	/* comment goes to end of line */
			if (curr_ch == '\r' || curr_ch == '\n' ||
			    curr_ch == EOF)
				state = state_start;
			break;
		case state_note:
			if (curr_ch == EOF) {
				do_warning("unexpected EOF in note");
				curr_token = tok_eof;
				return curr_token;
			} else if (curr_ch == '\'' || buf_pos >= MAX_LEN) {
				buf[buf_pos] = 0;
				state = state_done;
				curr_token = tok_number;
				if ((strlen(buf) != 3) ||
				    !strchr("ABCDEFG", buf[0]) ||
				    !strchr("-#b", buf[1]) ||
				    !strchr("0123456789A", buf[2])) {
					do_warning("invalid note value: %s",buf);
					curr_token_num = 0;
				} else {
					static const int notevals[7] =
						{ 9, 11, 0, 2, 4, 5, 7 };
					int note = notevals[buf[0] - 'A'];
					int octave = buf[2] - '0';
					if (buf[2] == 'A') { octave = 10; }
					if (buf[1] == '#') { note++; }
					else if (buf[1] == 'b') { note--; }
					note += 12 * octave;
					if (note < 0) note = 0;
					if (note > 127) note = 127;
					curr_token_num = midi_freqtab[note];
				}
				break;
			} else {
				buf[buf_pos++] = curr_ch;
				break;
			}
		case state_done:
			return curr_token;
		}
		curr_ch = lex_getchar();
	}
}

char *
dump_token(void)
{
	static char buf[MAX_LEN+128];
	switch (curr_token) {
		case tok_error:
			return "error";
		case tok_eof:
			return "EOF";
		case tok_semicolon:
			return ";";
		case tok_lparen:
			return "(";
		case tok_rparen:
			return ")";
		case tok_lbracket:
			return "[";
		case tok_rbracket:
			return "]";
		case tok_lbrace:
			return "{";
		case tok_rbrace:
			return "}";
		case tok_comma:
			return ",";
		case tok_equals:
			return "=";
		case tok_colon:
			return ":";
		case tok_coloncolon:
			return "::";
		case tok_plus:
			return "+";
		case tok_minus:
			return "-";
		case tok_star:
			return "*";
		case tok_slash:
			return "/";
		case tok_symbol:
			sprintf(buf,"sym %s",curr_token_sym->name);
			return buf;
		case tok_number:
			sprintf(buf,"num %f",curr_token_num);
			return buf;
		case tok_string:
			sprintf(buf,"str ``%s''",curr_token_str);
			return buf;
		default:
			return "error";
	}
}
